@@ -18,10 +18,12 @@
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 #
+from __future__ import print_function
+
 import samba.getopt as options
 import ldb
 import logging
-import common
+from . import common
 import json
 
 from samba.auth import system_session
@@ -30,16 +32,23 @@ from samba.netcmd import (
     CommandError,
     Option,
     SuperCommand,
-    )
+)
+from samba.netcmd.common import attr_default
 from samba.samdb import SamDB
 from samba import drs_utils, nttime2string, dsdb
 from samba.dcerpc import drsuapi, misc
 from samba.join import join_clone
-from samba.ndr import ndr_unpack
-from samba.dcerpc import drsblobs
 from samba import colour
-import logging
 
+from samba.uptodateness import (
+    get_partition_maps,
+    get_utdv_edges,
+    get_utdv_distances,
+    get_utdv_summary,
+    get_kcc_and_dsas,
+)
+from samba.common import get_string
+from samba.samdb import get_default_backend_store
 
 def drsuapi_connect(ctx):
     '''make a DRSUAPI connection to the server'''
@@ -48,6 +57,7 @@ def drsuapi_connect(ctx):
     except Exception as e:
         raise CommandError("DRS connection to %s failed" % ctx.server, e)
 
+
 def samdb_connect(ctx):
     '''make a ldap connection to the server'''
     try:
@@ -57,6 +67,7 @@ def samdb_connect(ctx):
     except Exception as e:
         raise CommandError("LDAP connection to %s failed" % ctx.server, e)
 
+
 def drs_errmsg(werr):
     '''return "was successful" or an error string'''
     (ecode, estring) = werr
@@ -65,15 +76,6 @@ def drs_errmsg(werr):
     return "failed, result %u (%s)" % (ecode, estring)
 
 
-
-def attr_default(msg, attrname, default):
-    '''get an attribute from a ldap msg with a default'''
-    if attrname in msg:
-        return msg[attrname][0]
-    return default
-
-
-
 def drs_parse_ntds_dn(ntds_dn):
     '''parse a NTDS DN returning a site and server'''
     a = ntds_dn.split(',')
@@ -86,6 +88,7 @@ def drs_parse_ntds_dn(ntds_dn):
 
 DEFAULT_SHOWREPL_FORMAT = 'classic'
 
+
 class cmd_drs_showrepl(Command):
     """Show replication status."""
 
@@ -147,7 +150,7 @@ class cmd_drs_showrepl(Command):
                 raise
         try:
             (site, server) = drs_parse_ntds_dn(n.source_dsa_obj_dn)
-            d["DSA"] = "%s\%s" % (site, server)
+            d["DSA"] = "%s\\%s" % (site, server)
         except RuntimeError:
             pass
         return d
@@ -246,7 +249,6 @@ class cmd_drs_showrepl(Command):
 
         self.message(colour.c_GREEN("[ALL GOOD]"))
 
-
     def summary_output(self):
         return self.summary_output_handler("summary")
 
@@ -271,10 +273,10 @@ class cmd_drs_showrepl(Command):
 
         dsa_details = {
             "options": int(attr_default(ntds[0], "options", 0)),
-            "objectGUID": self.samdb.schema_format_value(
-                "objectGUID", ntds[0]["objectGUID"][0]),
-            "invocationId": self.samdb.schema_format_value(
-                "objectGUID", ntds[0]["invocationId"][0])
+            "objectGUID": get_string(self.samdb.schema_format_value(
+                "objectGUID", ntds[0]["objectGUID"][0])),
+            "invocationId": get_string(self.samdb.schema_format_value(
+                "objectGUID", ntds[0]["invocationId"][0]))
         }
 
         conn = self.samdb.search(base=ntds_dn, expression="(objectClass=nTDSConnection)")
@@ -283,13 +285,13 @@ class cmd_drs_showrepl(Command):
 
         conn_details = []
         for c in conn:
-            c_rdn, sep, c_server_dn = c['fromServer'][0].partition(',')
+            c_rdn, sep, c_server_dn = str(c['fromServer'][0]).partition(',')
             d = {
                 'name': str(c['name']),
-                'remote DN': c['fromServer'][0],
+                'remote DN': str(c['fromServer'][0]),
                 'options': int(attr_default(c, 'options', 0)),
-                'enabled': (attr_default(c, 'enabledConnection',
-                                         'TRUE').upper() == 'TRUE')
+                'enabled': (get_string(attr_default(c, 'enabledConnection',
+                                         'TRUE')).upper() == 'TRUE')
             }
 
             conn_details.append(d)
@@ -297,7 +299,7 @@ class cmd_drs_showrepl(Command):
                 c_server_res = self.samdb.search(base=c_server_dn,
                                                  scope=ldb.SCOPE_BASE,
                                                  attrs=["dnsHostName"])
-                d['dns name'] = c_server_res[0]["dnsHostName"][0]
+                d['dns name'] = str(c_server_res[0]["dnsHostName"][0])
             except ldb.LdbError as e:
                 (errno, _) = e.args
                 if errno == ldb.ERR_NO_SUCH_OBJECT:
@@ -431,7 +433,7 @@ class cmd_drs_replicate(Command):
         Option("--local-online", help="pull changes into the local database (destination DC is ignored) as a normal online replication", action="store_true"),
         Option("--async-op", help="use ASYNC_OP for the replication", action="store_true"),
         Option("--single-object", help="Replicate only the object specified, instead of the whole Naming Context (only with --local)", action="store_true"),
-        ]
+    ]
 
     def drs_local_replicate(self, SOURCE_DC, NC, full_sync=False,
                             single_object=False,
@@ -441,8 +443,10 @@ class cmd_drs_replicate(Command):
         self.server = SOURCE_DC
         drsuapi_connect(self)
 
+        # Override the default flag LDB_FLG_DONT_CREATE_DB
         self.local_samdb = SamDB(session_info=system_session(), url=None,
-                                 credentials=self.creds, lp=self.lp)
+                                 credentials=self.creds, lp=self.lp,
+                                 flags=0)
 
         self.samdb = SamDB(url="ldap://%s" % self.server,
                            session_info=system_session(),
@@ -549,7 +553,6 @@ class cmd_drs_replicate(Command):
         source_dsa_guid = msg[0]['objectGUID'][0]
         dsa_options = int(attr_default(msg, 'options', 0))
 
-
         req_options = 0
         if not (dsa_options & dsdb.DS_NTDSDSA_OPT_DISABLE_OUTBOUND_REPL):
             req_options |= drsuapi.DRSUAPI_DRS_WRIT_REP
@@ -574,7 +577,6 @@ class cmd_drs_replicate(Command):
             self.message("Replicate from %s to %s was successful." % (SOURCE_DC, DEST_DC))
 
 
-
 class cmd_drs_bind(Command):
     """Show DRS capabilities of a server."""
 
@@ -605,41 +607,41 @@ class cmd_drs_bind(Command):
         (info, handle) = self.drsuapi.DsBind(misc.GUID(drsuapi.DRSUAPI_DS_BIND_GUID), bind_info)
 
         optmap = [
-            ("DRSUAPI_SUPPORTED_EXTENSION_BASE",     "DRS_EXT_BASE"),
-            ("DRSUAPI_SUPPORTED_EXTENSION_ASYNC_REPLICATION",   "DRS_EXT_ASYNCREPL"),
-            ("DRSUAPI_SUPPORTED_EXTENSION_REMOVEAPI",    "DRS_EXT_REMOVEAPI"),
-            ("DRSUAPI_SUPPORTED_EXTENSION_MOVEREQ_V2",   "DRS_EXT_MOVEREQ_V2"),
-            ("DRSUAPI_SUPPORTED_EXTENSION_GETCHG_COMPRESS",   "DRS_EXT_GETCHG_DEFLATE"),
-            ("DRSUAPI_SUPPORTED_EXTENSION_DCINFO_V1",    "DRS_EXT_DCINFO_V1"),
-            ("DRSUAPI_SUPPORTED_EXTENSION_RESTORE_USN_OPTIMIZATION",   "DRS_EXT_RESTORE_USN_OPTIMIZATION"),
-            ("DRSUAPI_SUPPORTED_EXTENSION_ADDENTRY",    "DRS_EXT_ADDENTRY"),
-            ("DRSUAPI_SUPPORTED_EXTENSION_KCC_EXECUTE",   "DRS_EXT_KCC_EXECUTE"),
-            ("DRSUAPI_SUPPORTED_EXTENSION_ADDENTRY_V2",   "DRS_EXT_ADDENTRY_V2"),
-            ("DRSUAPI_SUPPORTED_EXTENSION_LINKED_VALUE_REPLICATION",   "DRS_EXT_LINKED_VALUE_REPLICATION"),
-            ("DRSUAPI_SUPPORTED_EXTENSION_DCINFO_V2",    "DRS_EXT_DCINFO_V2"),
-            ("DRSUAPI_SUPPORTED_EXTENSION_INSTANCE_TYPE_NOT_REQ_ON_MOD","DRS_EXT_INSTANCE_TYPE_NOT_REQ_ON_MOD"),
-            ("DRSUAPI_SUPPORTED_EXTENSION_CRYPTO_BIND",   "DRS_EXT_CRYPTO_BIND"),
-            ("DRSUAPI_SUPPORTED_EXTENSION_GET_REPL_INFO",   "DRS_EXT_GET_REPL_INFO"),
-            ("DRSUAPI_SUPPORTED_EXTENSION_STRONG_ENCRYPTION",   "DRS_EXT_STRONG_ENCRYPTION"),
-            ("DRSUAPI_SUPPORTED_EXTENSION_DCINFO_V01",   "DRS_EXT_DCINFO_VFFFFFFFF"),
-            ("DRSUAPI_SUPPORTED_EXTENSION_TRANSITIVE_MEMBERSHIP",  "DRS_EXT_TRANSITIVE_MEMBERSHIP"),
-            ("DRSUAPI_SUPPORTED_EXTENSION_ADD_SID_HISTORY",   "DRS_EXT_ADD_SID_HISTORY"),
-            ("DRSUAPI_SUPPORTED_EXTENSION_POST_BETA3",   "DRS_EXT_POST_BETA3"),
-            ("DRSUAPI_SUPPORTED_EXTENSION_GETCHGREQ_V5",   "DRS_EXT_GETCHGREQ_V5"),
-            ("DRSUAPI_SUPPORTED_EXTENSION_GET_MEMBERSHIPS2",   "DRS_EXT_GETMEMBERSHIPS2"),
-            ("DRSUAPI_SUPPORTED_EXTENSION_GETCHGREQ_V6",   "DRS_EXT_GETCHGREQ_V6"),
-            ("DRSUAPI_SUPPORTED_EXTENSION_NONDOMAIN_NCS",   "DRS_EXT_NONDOMAIN_NCS"),
-            ("DRSUAPI_SUPPORTED_EXTENSION_GETCHGREQ_V8",   "DRS_EXT_GETCHGREQ_V8"),
-            ("DRSUAPI_SUPPORTED_EXTENSION_GETCHGREPLY_V5",   "DRS_EXT_GETCHGREPLY_V5"),
-            ("DRSUAPI_SUPPORTED_EXTENSION_GETCHGREPLY_V6",   "DRS_EXT_GETCHGREPLY_V6"),
-            ("DRSUAPI_SUPPORTED_EXTENSION_ADDENTRYREPLY_V3",   "DRS_EXT_WHISTLER_BETA3"),
-            ("DRSUAPI_SUPPORTED_EXTENSION_GETCHGREPLY_V7",   "DRS_EXT_WHISTLER_BETA3"),
-            ("DRSUAPI_SUPPORTED_EXTENSION_VERIFY_OBJECT",   "DRS_EXT_WHISTLER_BETA3"),
-            ("DRSUAPI_SUPPORTED_EXTENSION_XPRESS_COMPRESS",   "DRS_EXT_W2K3_DEFLATE"),
-            ("DRSUAPI_SUPPORTED_EXTENSION_GETCHGREQ_V10",   "DRS_EXT_GETCHGREQ_V10"),
-            ("DRSUAPI_SUPPORTED_EXTENSION_RESERVED_PART2",   "DRS_EXT_RESERVED_FOR_WIN2K_OR_DOTNET_PART2"),
+            ("DRSUAPI_SUPPORTED_EXTENSION_BASE", "DRS_EXT_BASE"),
+            ("DRSUAPI_SUPPORTED_EXTENSION_ASYNC_REPLICATION", "DRS_EXT_ASYNCREPL"),
+            ("DRSUAPI_SUPPORTED_EXTENSION_REMOVEAPI", "DRS_EXT_REMOVEAPI"),
+            ("DRSUAPI_SUPPORTED_EXTENSION_MOVEREQ_V2", "DRS_EXT_MOVEREQ_V2"),
+            ("DRSUAPI_SUPPORTED_EXTENSION_GETCHG_COMPRESS", "DRS_EXT_GETCHG_DEFLATE"),
+            ("DRSUAPI_SUPPORTED_EXTENSION_DCINFO_V1", "DRS_EXT_DCINFO_V1"),
+            ("DRSUAPI_SUPPORTED_EXTENSION_RESTORE_USN_OPTIMIZATION", "DRS_EXT_RESTORE_USN_OPTIMIZATION"),
+            ("DRSUAPI_SUPPORTED_EXTENSION_ADDENTRY", "DRS_EXT_ADDENTRY"),
+            ("DRSUAPI_SUPPORTED_EXTENSION_KCC_EXECUTE", "DRS_EXT_KCC_EXECUTE"),
+            ("DRSUAPI_SUPPORTED_EXTENSION_ADDENTRY_V2", "DRS_EXT_ADDENTRY_V2"),
+            ("DRSUAPI_SUPPORTED_EXTENSION_LINKED_VALUE_REPLICATION", "DRS_EXT_LINKED_VALUE_REPLICATION"),
+            ("DRSUAPI_SUPPORTED_EXTENSION_DCINFO_V2", "DRS_EXT_DCINFO_V2"),
+            ("DRSUAPI_SUPPORTED_EXTENSION_INSTANCE_TYPE_NOT_REQ_ON_MOD", "DRS_EXT_INSTANCE_TYPE_NOT_REQ_ON_MOD"),
+            ("DRSUAPI_SUPPORTED_EXTENSION_CRYPTO_BIND", "DRS_EXT_CRYPTO_BIND"),
+            ("DRSUAPI_SUPPORTED_EXTENSION_GET_REPL_INFO", "DRS_EXT_GET_REPL_INFO"),
+            ("DRSUAPI_SUPPORTED_EXTENSION_STRONG_ENCRYPTION", "DRS_EXT_STRONG_ENCRYPTION"),
+            ("DRSUAPI_SUPPORTED_EXTENSION_DCINFO_V01", "DRS_EXT_DCINFO_VFFFFFFFF"),
+            ("DRSUAPI_SUPPORTED_EXTENSION_TRANSITIVE_MEMBERSHIP", "DRS_EXT_TRANSITIVE_MEMBERSHIP"),
+            ("DRSUAPI_SUPPORTED_EXTENSION_ADD_SID_HISTORY", "DRS_EXT_ADD_SID_HISTORY"),
+            ("DRSUAPI_SUPPORTED_EXTENSION_POST_BETA3", "DRS_EXT_POST_BETA3"),
+            ("DRSUAPI_SUPPORTED_EXTENSION_GETCHGREQ_V5", "DRS_EXT_GETCHGREQ_V5"),
+            ("DRSUAPI_SUPPORTED_EXTENSION_GET_MEMBERSHIPS2", "DRS_EXT_GETMEMBERSHIPS2"),
+            ("DRSUAPI_SUPPORTED_EXTENSION_GETCHGREQ_V6", "DRS_EXT_GETCHGREQ_V6"),
+            ("DRSUAPI_SUPPORTED_EXTENSION_NONDOMAIN_NCS", "DRS_EXT_NONDOMAIN_NCS"),
+            ("DRSUAPI_SUPPORTED_EXTENSION_GETCHGREQ_V8", "DRS_EXT_GETCHGREQ_V8"),
+            ("DRSUAPI_SUPPORTED_EXTENSION_GETCHGREPLY_V5", "DRS_EXT_GETCHGREPLY_V5"),
+            ("DRSUAPI_SUPPORTED_EXTENSION_GETCHGREPLY_V6", "DRS_EXT_GETCHGREPLY_V6"),
+            ("DRSUAPI_SUPPORTED_EXTENSION_ADDENTRYREPLY_V3", "DRS_EXT_WHISTLER_BETA3"),
+            ("DRSUAPI_SUPPORTED_EXTENSION_GETCHGREPLY_V7", "DRS_EXT_WHISTLER_BETA3"),
+            ("DRSUAPI_SUPPORTED_EXTENSION_VERIFY_OBJECT", "DRS_EXT_WHISTLER_BETA3"),
+            ("DRSUAPI_SUPPORTED_EXTENSION_XPRESS_COMPRESS", "DRS_EXT_W2K3_DEFLATE"),
+            ("DRSUAPI_SUPPORTED_EXTENSION_GETCHGREQ_V10", "DRS_EXT_GETCHGREQ_V10"),
+            ("DRSUAPI_SUPPORTED_EXTENSION_RESERVED_PART2", "DRS_EXT_RESERVED_FOR_WIN2K_OR_DOTNET_PART2"),
             ("DRSUAPI_SUPPORTED_EXTENSION_RESERVED_PART3", "DRS_EXT_RESERVED_FOR_WIN2K_OR_DOTNET_PART3")
-            ]
+        ]
 
         optmap_ext = [
             ("DRSUAPI_SUPPORTED_EXTENSION_ADAM", "DRS_EXT_ADAM"),
@@ -672,7 +674,6 @@ class cmd_drs_bind(Command):
             self.message("Forest GUID: %s" % info.info.config_dn_guid)
 
 
-
 class cmd_drs_options(Command):
     """Query or change 'options' for NTDS Settings object of a Domain Controller."""
 
@@ -688,8 +689,8 @@ class cmd_drs_options(Command):
 
     takes_options = [
         Option("--dsa-option", help="DSA option to enable/disable", type="str",
-               metavar="{+|-}IS_GC | {+|-}DISABLE_INBOUND_REPL | {+|-}DISABLE_OUTBOUND_REPL | {+|-}DISABLE_NTDSCONN_XLATE" ),
-        ]
+               metavar="{+|-}IS_GC | {+|-}DISABLE_INBOUND_REPL | {+|-}DISABLE_OUTBOUND_REPL | {+|-}DISABLE_NTDSCONN_XLATE"),
+    ]
 
     option_map = {"IS_GC": 0x00000001,
                   "DISABLE_INBOUND_REPL": 0x00000002,
@@ -726,10 +727,10 @@ class cmd_drs_options(Command):
                 dsa_opts |= self.option_map[flag]
             else:
                 dsa_opts &= ~self.option_map[flag]
-            #save new options
+            # save new options
             m = ldb.Message()
             m.dn = ldb.Dn(self.samdb, ntds_dn)
-            m["options"]= ldb.MessageElement(str(dsa_opts), ldb.FLAG_MOD_REPLACE, "options")
+            m["options"] = ldb.MessageElement(str(dsa_opts), ldb.FLAG_MOD_REPLACE, "options")
             self.samdb.modify(m)
             # print out new DSA options
             cur_opts = [x for x in self.option_map if self.option_map[x] & dsa_opts]
@@ -752,31 +753,116 @@ class cmd_drs_clone_dc_database(Command):
         Option("--targetdir", help="where to store provision (required)", type=str),
         Option("-q", "--quiet", help="Be quiet", action="store_true"),
         Option("--include-secrets", help="Also replicate secret values", action="store_true"),
-        Option("-v", "--verbose", help="Be verbose", action="store_true")
-       ]
+        Option("--backend-store", type="choice", metavar="BACKENDSTORE",
+               choices=["tdb", "mdb"],
+               help="Specify the database backend to be used "
+                    "(default is %s)" % get_default_backend_store()),
+        Option("--backend-store-size", type="bytes", metavar="SIZE",
+               help="Specify the size of the backend database, currently" +
+                    "only supported by lmdb backends (default is 8 Gb).")
+    ]
 
     takes_args = ["domain"]
 
     def run(self, domain, sambaopts=None, credopts=None,
             versionopts=None, server=None, targetdir=None,
-            quiet=False, verbose=False, include_secrets=False):
+            quiet=False, verbose=False, include_secrets=False,
+            backend_store=None, backend_store_size=None):
         lp = sambaopts.get_loadparm()
         creds = credopts.get_credentials(lp)
 
-        logger = self.get_logger()
-        if verbose:
-            logger.setLevel(logging.DEBUG)
-        elif quiet:
-            logger.setLevel(logging.WARNING)
-        else:
-            logger.setLevel(logging.INFO)
+        logger = self.get_logger(verbose=verbose, quiet=quiet)
 
         if targetdir is None:
             raise CommandError("--targetdir option must be specified")
 
+        join_clone(logger=logger, server=server, creds=creds, lp=lp,
+                   domain=domain, dns_backend='SAMBA_INTERNAL',
+                   targetdir=targetdir, include_secrets=include_secrets,
+                   backend_store=backend_store,
+                   backend_store_size=backend_store_size)
 
-        join_clone(logger=logger, server=server, creds=creds, lp=lp, domain=domain,
-                   targetdir=targetdir, include_secrets=include_secrets)
+
+class cmd_drs_uptodateness(Command):
+    """Show uptodateness status"""
+
+    synopsis = "%prog [options]"
+
+    takes_optiongroups = {
+        "sambaopts": options.SambaOptions,
+        "versionopts": options.VersionOptions,
+        "credopts": options.CredentialsOptions,
+    }
+
+    takes_options = [
+        Option("-H", "--URL", metavar="URL", dest="H",
+               help="LDB URL for database or target server"),
+        Option("-p", "--partition",
+               help="restrict to this partition"),
+        Option("--json", action='store_true',
+               help="Print data in json format"),
+        Option("--maximum", action='store_true',
+               help="Print maximum out-of-date-ness only"),
+        Option("--median", action='store_true',
+               help="Print median out-of-date-ness only"),
+        Option("--full", action='store_true',
+               help="Print full out-of-date-ness data"),
+    ]
+
+    def format_as_json(self, partitions_summaries):
+        return json.dumps(partitions_summaries, indent=2)
+
+    def format_as_text(self, partitions_summaries):
+        lines = []
+        for part_name, summary in partitions_summaries.items():
+            items = ['%s: %s' % (k, v) for k, v in summary.items()]
+            line = '%-15s %s' % (part_name, '  '.join(items))
+            lines.append(line)
+        return '\n'.join(lines)
+
+    def run(self, H=None, partition=None,
+            json=False, maximum=False, median=False, full=False,
+            sambaopts=None, credopts=None, versionopts=None,
+            quiet=False, verbose=False):
+
+        lp = sambaopts.get_loadparm()
+        creds = credopts.get_credentials(lp, fallback_machine=True)
+        local_kcc, dsas = get_kcc_and_dsas(H, lp, creds)
+        samdb = local_kcc.samdb
+        short_partitions, _ = get_partition_maps(samdb)
+        if partition:
+            if partition in short_partitions:
+                part_dn = short_partitions[partition]
+                # narrow down to specified partition only
+                short_partitions = {partition: part_dn}
+            else:
+                raise CommandError("unknown partition %s" % partition)
+
+        filters = []
+        if maximum:
+            filters.append('maximum')
+        if median:
+            filters.append('median')
+
+        partitions_distances = {}
+        partitions_summaries = {}
+        for part_name, part_dn in short_partitions.items():
+            utdv_edges = get_utdv_edges(local_kcc, dsas, part_dn, lp, creds)
+            distances = get_utdv_distances(utdv_edges, dsas)
+            summary = get_utdv_summary(distances, filters=filters)
+            partitions_distances[part_name] = distances
+            partitions_summaries[part_name] = summary
+
+        if full:
+            # always print json format
+            output = self.format_as_json(partitions_distances)
+        else:
+            if json:
+                output = self.format_as_json(partitions_summaries)
+            else:
+                output = self.format_as_text(partitions_summaries)
+
+        print(output, file=self.outf)
 
 
 class cmd_drs(SuperCommand):
@@ -789,3 +875,4 @@ class cmd_drs(SuperCommand):
     subcommands["showrepl"] = cmd_drs_showrepl()
     subcommands["options"] = cmd_drs_options()
     subcommands["clone-dc-database"] = cmd_drs_clone_dc_database()
+    subcommands["uptodateness"] = cmd_drs_uptodateness()
